/******************************************************************
* FILENAME: utils.cpp
*
* DESCRIPTION:
*	functions used by instvc.dll
*
* REFERENCES:

	Much of the security source code here was obtained from Michael Nelson.
	I am on his 'drop list' for source code enhancements to the
	DCOMPERM com object he has developed.
*

* NOTES:

*******************************************************************/

#define	_WIN32_DCOM
#include <windows.h>
#include <comcat.h>
#include <comdef.h>
#include <stdio.h>
#include <conio.h>
#include <tchar.h>
#include <lm.h>

#include "ntsecapi.h"

#define REPORT_CALL_FAILURES
#include "utils.h"

inline
const TCHAR *FormatLastError(DWORD lastError, TCHAR *formatBuffer, const TCHAR *caption)
{
	TCHAR errorText[4096];

	if (!FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, lastError, 0, errorText, sizeof(errorText), 0))
		_tcscpy(errorText, __T("Unknown error"));

	_tcstok(errorText, __T("\n\r"));	//	null terminate
	_stprintf(formatBuffer, __T("%s, \"%s\" (0x%x)\n"), caption, errorText, lastError);

	return formatBuffer;
}

DWORD SetAccountRights( LPWSTR User, LPWSTR Privilege )
{
    LSA_HANDLE            policyHandle;
    LSA_OBJECT_ATTRIBUTES objectAttributes;
    PSID                  principalSID;
    LSA_UNICODE_STRING    lsaPrivilegeString;
    DWORD                 dw;
    SERVER_INFO_101       *si101;
    USER_MODALS_INFO_2    *umi2;
    LPWSTR                domainController;
    LSA_UNICODE_STRING    lsaDomainController;
    DWORD                 returnValue;

    domainController = NULL;
    memset (&objectAttributes, 0, sizeof (LSA_OBJECT_ATTRIBUTES));

    //
    // A whole bunch of crufty code to make sure that if we're on
    // a BDC that we go to the PDC to adjust the account rights there.
    //

    if ((dw = NetServerGetInfo (NULL, 101, (LPBYTE *) &si101)) == 0)
    {
        if (si101->sv101_type & SV_TYPE_DOMAIN_BAKCTRL)
        {
            if ((dw = NetUserModalsGet (NULL, 2, (LPBYTE *) &umi2)) == 0)
            {
                NetGetDCName (NULL, umi2->usrmod2_domain_name, (LPBYTE *) &domainController);

                lsaDomainController.Length = (USHORT) (wcslen (domainController) * sizeof (WCHAR));
                lsaDomainController.MaximumLength = (USHORT) (lsaDomainController.Length + sizeof (WCHAR));
                lsaDomainController.Buffer = &domainController[2];

                if (LsaOpenPolicy (&lsaDomainController,
                                   &objectAttributes,
                                   POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES,
                                   &policyHandle) != ERROR_SUCCESS)
                {
                    NetApiBufferFree (domainController);
                    NetApiBufferFree (umi2);
                    NetApiBufferFree (si101);

					ReportFailure("SetAccountRights", "LsaOpenPolicy", __FILE__, __LINE__);

                    return HRESULT_FROM_WIN32 (GetLastError());
                }

                NetApiBufferFree (domainController);
                NetApiBufferFree (umi2);
                NetApiBufferFree (si101);
            }
        } else
        {
            NetApiBufferFree (si101);
        }
    }

    if (!domainController)
    {
        memset (&objectAttributes, 0, sizeof (LSA_OBJECT_ATTRIBUTES));
        if (LsaOpenPolicy (NULL,
                           &objectAttributes,
                           POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES,
                           &policyHandle) != ERROR_SUCCESS)
        {
			ReportFailure("SetAccountRights", "LsaOpenPolicy", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (GetLastError());
        }
    }

    lsaPrivilegeString.Length = (USHORT) (wcslen (Privilege) * sizeof (WCHAR));
    lsaPrivilegeString.MaximumLength = (USHORT) (lsaPrivilegeString.Length + sizeof (WCHAR));
    lsaPrivilegeString.Buffer = Privilege;

    returnValue = GetPrincipalSID (User, &principalSID);
    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetAccountRights", "GetPrincipalSID", __FILE__, __LINE__);
        return returnValue;
	}

    if (LsaAddAccountRights (policyHandle,
                             principalSID,
                             &lsaPrivilegeString,
                             1) != ERROR_SUCCESS)
    {
        free (principalSID);
        LsaClose (policyHandle);
		ReportFailure("SetAccountRights", "LsaAddAccountRights", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    free (principalSID);
    LsaClose (policyHandle);

    return ERROR_SUCCESS;
}

void InitLsaString( PLSA_UNICODE_STRING LsaString, LPWSTR String )
{
    DWORD StringLength;

    if(String == NULL) {
        LsaString->Buffer = NULL;
        LsaString->Length = 0;
        LsaString->MaximumLength = 0;
        return;
    }

    StringLength = lstrlenW(String);
    LsaString->Buffer = String;
    LsaString->Length = (USHORT) StringLength * sizeof(WCHAR);
    LsaString->MaximumLength=(USHORT)(StringLength+1) * sizeof(WCHAR);
}


NTSTATUS SetPrivilegeOnAccount(
	LSA_HANDLE PolicyHandle,    // open policy handle
    PSID AccountSid,            // SID to grant privilege to
    LPWSTR PrivilegeName       // privilege to grant (Unicode)
    )
{
    LSA_UNICODE_STRING PrivilegeString;

    // Create a LSA_UNICODE_STRING for the privilege name.
    InitLsaString(&PrivilegeString, PrivilegeName);

    return LsaAddAccountRights(
                PolicyHandle,       // open policy handle
                AccountSid,         // target SID
                &PrivilegeString,   // privileges
                1                   // privilege count
                );
}

NTSTATUS OpenPolicy( DWORD DesiredAccess, PLSA_HANDLE PolicyHandle )
{
    LSA_OBJECT_ATTRIBUTES ObjectAttributes;

    ZeroMemory(&ObjectAttributes, sizeof(ObjectAttributes));

    return LsaOpenPolicy( NULL, &ObjectAttributes, DesiredAccess, PolicyHandle );
}


BOOL HasEXEExtension(const char *filename)
{
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];

	_splitpath(filename, drive, dir, fname, ext);
	
	return strstr(strlwr(ext), "exe")? true : false;
}

DWORD GetCurrentUserSID( PSID *Sid )
{
    TOKEN_USER  *tokenUser = NULL;
    HANDLE      tokenHandle;
    DWORD       tokenSize;
    DWORD       sidLength;

    if (OpenProcessToken (GetCurrentProcess(), TOKEN_QUERY, &tokenHandle))
    {
        GetTokenInformation (tokenHandle,
                             TokenUser,
                             NULL,
                             0,
                             &tokenSize);

        tokenUser = (TOKEN_USER *) malloc (tokenSize);

        if (GetTokenInformation (tokenHandle,
                                 TokenUser,
                                 tokenUser,
                                 tokenSize,
                                 &tokenSize))
        {
            sidLength = GetLengthSid (tokenUser->User.Sid);
            *Sid = (PSID) malloc (sidLength);

            memcpy (*Sid, tokenUser->User.Sid, sidLength);
            CloseHandle (tokenHandle);
        } else
        {
            free (tokenUser);
            return HRESULT_FROM_WIN32 (GetLastError());
        }
    } else
    {
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    if (tokenUser)
        free (tokenUser);

    return ERROR_SUCCESS;
}



DWORD GetPrincipalSID( LPWSTR Principal, PSID *Sid )
{
    DWORD        sidSize = 0;
    WCHAR        refDomain [256];
    DWORD        refDomainSize = 0;
    DWORD        returnValue;
    SID_NAME_USE snu;
    PSID         psidEveryone;
    SID_IDENTIFIER_AUTHORITY sepWorldSidAuthority = SECURITY_WORLD_SID_AUTHORITY;
    WCHAR        principal [256];


    wcscpy (principal, Principal);

    //
    // Chop of NT Authority\\
    //

    if (wcsnicmp (L"NT Authority\\", Principal, 13) == 0)
        wcscpy (principal, &Principal [13]);

    if (wcsicmp (L"Everyone", principal) == 0)
    {
        if (!AllocateAndInitializeSid (&sepWorldSidAuthority,
                                       1,
                                       SECURITY_WORLD_RID,
                                       0,
                                       0,
                                       0,
                                       0,
                                       0,
                                       0,
                                       0,
                                       &psidEveryone
                                       ))
        {
			ReportFailure("GetPrincipalSID", "AllocateAndInitializeSid", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (GetLastError());
        }

        //
        // Sid is big just in case
        //

        *Sid = (PSID) malloc (GetLengthSid (psidEveryone));
        if (!CopySid (GetLengthSid (psidEveryone), *Sid, psidEveryone))
        {
            FreeSid (psidEveryone);
			ReportFailure("GetPrincipalSID", "CopySid", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (GetLastError());
        }

        FreeSid (psidEveryone);

        return ERROR_SUCCESS;
    }

    LookupAccountNameW (NULL,
                       principal,
                       *Sid,
                       &sidSize,
                       refDomain,
                       &refDomainSize,
                       &snu);

    returnValue = HRESULT_FROM_WIN32 (GetLastError());
    if (returnValue != HRESULT_FROM_WIN32 (ERROR_INSUFFICIENT_BUFFER))
	{
		ReportFailure2("GetPrincipalSID", "LookupAccountName", __FILE__, __LINE__, returnValue );
        return returnValue;
	}

    *Sid = (PSID) malloc (sidSize);

    if (!LookupAccountNameW (NULL,
                            principal,
                            *Sid,
                            &sidSize,
                            refDomain,
                            &refDomainSize,
                            &snu))
    {
		ReportFailure("GetPrincipalSID", "LookupAccountName", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    return ERROR_SUCCESS;
}


LPTSTR SystemMessage( LPTSTR Buffer, HRESULT hr )
{
	LPTSTR	message;

	if (HRESULT_FACILITY(hr) == FACILITY_WINDOWS)
		hr = HRESULT_CODE(hr);
 
	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
					FORMAT_MESSAGE_FROM_SYSTEM,
					NULL,
					hr,
					MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					(LPTSTR) &message,
					0,
					NULL);

	wsprintf (Buffer, TEXT("%s(%lx)\n"), message, hr);
    
	LocalFree (message);
	return Buffer;
}

DWORD CreateNewSD( SECURITY_DESCRIPTOR **SD )
{
    DWORD   returnValue = ERROR_SUCCESS;

#ifndef WIN95
    PACL    dacl;
    DWORD   sidLength;
    PSID    sid;
    PSID    groupSID;
    PSID    ownerSID;

    *SD = NULL;

    returnValue = GetCurrentUserSID (&sid);
    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("CreateNewSD", "GetCurrentUserSID", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (returnValue);
	}

    sidLength = GetLengthSid (sid);

    *SD = (SECURITY_DESCRIPTOR *) malloc (
        (sizeof (ACL)+sizeof (ACCESS_ALLOWED_ACE)+sidLength) +
        (2 * sidLength) +
        sizeof (SECURITY_DESCRIPTOR));

    memset (*SD, 0xfe, sizeof (DWORD));

    groupSID = (SID *) (*SD + 1);   // XXX
    ownerSID = (SID *) (((BYTE *) groupSID) + sidLength);

    // TEMPORARY
    dacl = (PACL) malloc (8192);
    // dacl = (ACL *) (((BYTE *) ownerSID) + sidLength);

    try
    {
        if (!InitializeSecurityDescriptor (*SD, SECURITY_DESCRIPTOR_REVISION))
            throw;

        if (!InitializeAcl (dacl,
                            sizeof (ACL)+sizeof (ACCESS_ALLOWED_ACE)+sidLength,
                            ACL_REVISION2))
            throw;



        if (!AddAccessAllowedAce (dacl,
                                  ACL_REVISION2,
                                  COM_RIGHTS_EXECUTE,
                                  sid))
            throw;

        if (!SetSecurityDescriptorDacl (*SD, TRUE, dacl, FALSE))
            throw;

        memcpy (groupSID, sid, sidLength);
        if (!SetSecurityDescriptorGroup (*SD, groupSID, FALSE))
            throw;

        memcpy (ownerSID, sid, sidLength);
        if (!SetSecurityDescriptorOwner (*SD, ownerSID, FALSE))
            throw;
    } catch (...)
    {
        returnValue = HRESULT_FROM_WIN32 (GetLastError());
    }

    // free (dacl); // XXX
    free (sid);
#endif

    return returnValue;
}



DWORD MakeSDAbsolute( PSECURITY_DESCRIPTOR OldSD, PSECURITY_DESCRIPTOR *NewSD )
{
#ifndef WIN95
    PSECURITY_DESCRIPTOR  sd;
    DWORD                 descriptorSize;
    DWORD                 daclSize;
    DWORD                 saclSize;
    DWORD                 ownerSIDSize;
    DWORD                 groupSIDSize;
    PACL                  dacl;
    PACL                  sacl;
    PSID                  ownerSID;
    PSID                  groupSID;
    BOOL                  present;
    BOOL                  systemDefault;

    //
    // Get SACL
    //

    if (!GetSecurityDescriptorSacl (OldSD, &present, &sacl, &systemDefault))
	{
		ReportFailure("MakeSDAbsolute", "GetSecurityDescriptorSacl", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
	}

    if (sacl && present)
    {
        saclSize = sacl->AclSize;
    } else saclSize = 0;

    //
    // Get DACL
    //

    if (!GetSecurityDescriptorDacl (OldSD, &present, &dacl, &systemDefault))
	{
		ReportFailure("MakeSDAbsolute", "GetSecurityDescriptorDacl", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
	}

    if (dacl && present)
    {
        daclSize = dacl->AclSize;
    } else daclSize = 0;

    //
    // Get Owner
    //

    if (!GetSecurityDescriptorOwner (OldSD, &ownerSID, &systemDefault))
	{
		ReportFailure("MakeSDAbsolute", "GetSecurityDescriptorOwner", __FILE__, __LINE__);     
		return HRESULT_FROM_WIN32 (GetLastError());
	}

    ownerSIDSize = GetLengthSid (ownerSID);

    //
    // Get Group
    //

    if (!GetSecurityDescriptorGroup (OldSD, &groupSID, &systemDefault))
	{
		ReportFailure("MakeSDAbsolute", "GetSecurityDescriptorGroup", __FILE__, __LINE__);     
		return HRESULT_FROM_WIN32 (GetLastError());
	}

    groupSIDSize = GetLengthSid (groupSID);

    //
    // Do the conversion
    //

    descriptorSize = 0;

    MakeAbsoluteSD (OldSD, NULL, &descriptorSize, dacl, &daclSize, sacl,
                    &saclSize, ownerSID, &ownerSIDSize, groupSID,
                    &groupSIDSize);

    sd = (PSECURITY_DESCRIPTOR) malloc (SECURITY_DESCRIPTOR_MIN_LENGTH);
    if (!InitializeSecurityDescriptor (sd, SECURITY_DESCRIPTOR_REVISION))
	{
		ReportFailure("MakeSDAbsolute", "InitializeSecurityDescriptor", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
	}

    if (!MakeAbsoluteSD (OldSD, sd, &descriptorSize, dacl, &daclSize, sacl,
                         &saclSize, ownerSID, &ownerSIDSize, groupSID,
                         &groupSIDSize))
	{
		ReportFailure("MakeSDAbsolute", "MakeAbsoluteSD", __FILE__, __LINE__);     
		return HRESULT_FROM_WIN32 (GetLastError());
	}

    *NewSD = sd;
#endif
    return ERROR_SUCCESS;
}

DWORD SetNamedValueSD( HKEY RootKey, LPWSTR KeyName, LPWSTR ValueName, SECURITY_DESCRIPTOR *SD )
{
#ifndef WIN95
    DWORD   returnValue;
    DWORD   disposition;
    HKEY    registryKey;

    //
    // Create new key or open existing key
    //

    returnValue = RegCreateKeyExW (RootKey, KeyName, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &registryKey, &disposition);
    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetNamedValueSD", "RegCreateKeyExW", __FILE__, __LINE__);     
		return HRESULT_FROM_WIN32 (returnValue);
	}

    //
    // Write the security descriptor
    //

    returnValue = RegSetValueExW (registryKey, ValueName, 0, REG_BINARY, (LPBYTE) SD, GetSecurityDescriptorLength (SD));
    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetNamedValueSD", "RegSetValueExW", __FILE__, __LINE__);     
        return HRESULT_FROM_WIN32 (returnValue);
	}

    RegCloseKey (registryKey);
#endif
    return ERROR_SUCCESS;
}

DWORD GetNamedValueSD( HKEY RootKey, LPWSTR KeyName, LPWSTR ValueName, SECURITY_DESCRIPTOR **SD, BOOL *NewSD )
{
#ifndef WIN95
    DWORD               returnValue;
    HKEY                registryKey;
    DWORD               valueType;
    DWORD               valueSize;

    *NewSD = FALSE;

    //
    // Get the security descriptor from the named value. If it doesn't
    // exist, create a fresh one.
    //

    returnValue = RegOpenKeyExW (RootKey, KeyName, 0, KEY_ALL_ACCESS, &registryKey);

    if (returnValue != ERROR_SUCCESS)
    {
        if (returnValue == ERROR_FILE_NOT_FOUND)
        {
            *SD = NULL;
            returnValue = CreateNewSD (SD);

            if (returnValue != ERROR_SUCCESS)
			{
				ReportFailure("GetNamedValueSD", "CreateNewSD", __FILE__, __LINE__);
                return HRESULT_FROM_WIN32 (returnValue);
			}

            *NewSD = TRUE;
            return ERROR_SUCCESS;
        } else
		{
			ReportFailure("GetNamedValueSD", "RegOpenKeyExW", __FILE__, __LINE__);
			return HRESULT_FROM_WIN32 (returnValue);
		}
    }

    returnValue = RegQueryValueExW (registryKey, ValueName, NULL, &valueType, NULL, &valueSize);

    if (returnValue && returnValue != ERROR_INSUFFICIENT_BUFFER)
    {
        *SD = NULL;
        returnValue = CreateNewSD (SD);
        if (returnValue != ERROR_SUCCESS)
		{
			ReportFailure("GetNamedValueSD", "CreateNewSD", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (returnValue);
		}

        *NewSD = TRUE;
    } else
    {
        *SD = (SECURITY_DESCRIPTOR *) malloc (valueSize);

        returnValue = RegQueryValueExW (registryKey, ValueName, NULL, &valueType, (LPBYTE) *SD, &valueSize);
        if (returnValue)
        {
            free (*SD);

            *SD = NULL;
            returnValue = CreateNewSD (SD);
            if (returnValue != ERROR_SUCCESS)
			{
				ReportFailure("GetNamedValueSD", "CreateNewSD", __FILE__, __LINE__);
                return HRESULT_FROM_WIN32 (returnValue);
			}

            *NewSD = TRUE;
        }
    }

    RegCloseKey (registryKey);
#endif

    return ERROR_SUCCESS;
}

DWORD AddPrincipalToNamedValueSD( HKEY RootKey, LPWSTR KeyName, LPWSTR ValueName, LPWSTR Principal, BOOL Permit )
{
    DWORD               returnValue;
    SECURITY_DESCRIPTOR *sd;
    SECURITY_DESCRIPTOR *sdSelfRelative;
    SECURITY_DESCRIPTOR *sdAbsolute;
    DWORD               secDescSize;
    BOOL                present;
    BOOL                defaultDACL;
    PACL                dacl;
    BOOL                newSD = FALSE;


    returnValue = GetNamedValueSD (RootKey, KeyName, ValueName, &sd, &newSD);

    //
    // Get security descriptor from registry or create a new one
    //

    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("AddPrincipalToNamedValueSD", "GetNamedValueSD", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (returnValue);
	}

    if (!GetSecurityDescriptorDacl (sd, &present, &dacl, &defaultDACL))
	{
		ReportFailure("AddPrincipalToNamedValueSD", "GetSecurityDescriptorDacl", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
	}


// ** TEMPORARY **

//
// throw dacl into our own private buffer so that freeing semantics of
// AddAccessAllowedACEToACL don't screw us up. To fix in a future release.
//

    ACL_SIZE_INFORMATION aclSizeInfo;
    GetAclInformation (dacl, (LPVOID) &aclSizeInfo, (DWORD) sizeof (ACL_SIZE_INFORMATION), AclSizeInformation);
    PACL daclx;
    daclx = (PACL) malloc (aclSizeInfo.AclBytesInUse);
    memcpy (daclx, dacl, aclSizeInfo.AclBytesInUse);

// ** TEMPORARY **

    if (!((wcsicmp (Principal, L"SYSTEM") == 0) ||
         (wcsicmp (Principal, L"NT Authority\\SYSTEM") == 0)))
    {
        // need to check retval on aaaata

        if (!IsPrincipalInACL (daclx, L"NT Authority\\SYSTEM"))
            AddAccessAllowedACEToACL (&daclx, COM_RIGHTS_EXECUTE, L"NT Authority\\SYSTEM");
    }

    //
    // Add the Principal that the caller wants added
    //

    if (Permit)
        returnValue = AddAccessAllowedACEToACL (&daclx, COM_RIGHTS_EXECUTE, Principal); else
        returnValue = AddAccessDeniedACEToACL (&daclx, COM_RIGHTS_EXECUTE, Principal);

    if (returnValue != ERROR_SUCCESS)
    {
        free (sd);
		ReportFailure("AddPrincipalToNamedValueSD", "AddAccess???ACEToACL", __FILE__, __LINE__);
        return returnValue;
    }

    //
    // Make the security descriptor absolute if it isn't new
    //

    if (!newSD)
        MakeSDAbsolute ((PSECURITY_DESCRIPTOR) sd, (PSECURITY_DESCRIPTOR *) &sdAbsolute); else
        sdAbsolute = sd;

    //
    // Set the discretionary ACL on the security descriptor
    //

    if (!SetSecurityDescriptorDacl (sdAbsolute, TRUE, daclx, FALSE))
    {
        if (!newSD)
            free (sd);
        free (sdAbsolute);
		ReportFailure("AddPrincipalToNamedValueSD", "SetSecurityDescriptorDacl", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    //
    // Make the security descriptor self-relative so that we can
    // store it in the registry
    //

    secDescSize = 0;
    MakeSelfRelativeSD (sdAbsolute, NULL, &secDescSize);
    sdSelfRelative = (SECURITY_DESCRIPTOR *) malloc (secDescSize);
    if (!MakeSelfRelativeSD (sdAbsolute, sdSelfRelative, &secDescSize))
    {
        if (!newSD)
            free (sd);
        free (sdAbsolute);

//
// XXX: if someone does heapfree or something, this may not work right
//
		ReportFailure("AddPrincipalToNamedValueSD", "MakeSelfRelativeSD", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    //
    // Store the security descriptor in the registry
    //

    SetNamedValueSD (RootKey, KeyName, ValueName, sdSelfRelative);

    if (!newSD)
        free (sd);

    free (daclx); // XXX
    free (sdSelfRelative);
    free (sdAbsolute);

    return ERROR_SUCCESS;
}

DWORD RemovePrincipalFromNamedValueSD( HKEY RootKey, LPWSTR KeyName, LPWSTR ValueName, LPWSTR Principal )
{
    DWORD               returnValue;
    SECURITY_DESCRIPTOR *sd;
    SECURITY_DESCRIPTOR *sdSelfRelative;
    SECURITY_DESCRIPTOR *sdAbsolute;
    DWORD               secDescSize;
    BOOL                present;
    BOOL                defaultDACL;
    PACL                dacl;
    BOOL                newSD = FALSE;

    returnValue = GetNamedValueSD (RootKey, KeyName, ValueName, &sd, &newSD);

    //
    // Get security descriptor from registry
    //

    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("RemovePrincipalFromNamedValueSD", "GetNamedValueSD", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (returnValue);
	}

    if (!GetSecurityDescriptorDacl (sd, &present, &dacl, &defaultDACL))
	{
		ReportFailure("RemovePrincipalFromNamedValueSD", "GetSecurityDescriptorDacl", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
	}

    //
    // If the security descriptor is new, add the required Principals to it
    //

    if (newSD)
    {
        return ERROR_SUCCESS;
    }

    //
    // Remove the Principal that the caller wants removed
    //

    returnValue = RemovePrincipalFromACL (dacl, Principal);
    if (returnValue != ERROR_SUCCESS)
    {
		ReportFailure("RemovePrincipalFromNamedValueSD", "RemovePrincipalFromACL", __FILE__, __LINE__);

        free (sd);
        return HRESULT_FROM_WIN32 (returnValue);
    }

    //
    // Make the security descriptor absolute if it isn't new
    //

    if (!newSD)
        MakeSDAbsolute ((PSECURITY_DESCRIPTOR) sd, (PSECURITY_DESCRIPTOR *) &sdAbsolute);
	else
        sdAbsolute = sd;

    //
    // Set the discretionary ACL on the security descriptor
    //

    if (!SetSecurityDescriptorDacl (sdAbsolute, TRUE, dacl, FALSE))
    {
		ReportFailure("RemovePrincipalFromNamedValueSD", "SetSecurityDescriptorDacl", __FILE__, __LINE__);
        if (!newSD)
            free (sd);

        free (sdAbsolute);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    //
    // Make the security descriptor self-relative so that we can
    // store it in the registry
    //

    secDescSize = 0;
    MakeSelfRelativeSD (sdAbsolute, NULL, &secDescSize);
    sdSelfRelative = (SECURITY_DESCRIPTOR *) malloc (secDescSize);
    if (!MakeSelfRelativeSD (sdAbsolute, sdSelfRelative, &secDescSize))
    {
		ReportFailure("RemovePrincipalFromNamedValueSD", "MakeSelfRelativeSD", __FILE__, __LINE__);
        if (!newSD)
            free (sd);

        free (sdAbsolute);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    //
    // Store the security descriptor in the registry
    //

    SetNamedValueSD (RootKey, KeyName, ValueName, sdSelfRelative);

    if (!newSD)
        free (sd);

    free (sdAbsolute);
    free (sdSelfRelative);

    return ERROR_SUCCESS;
}


DWORD CopyACL( PACL OldACL, PACL NewACL )
{
    ACL_SIZE_INFORMATION  aclSizeInfo;
    LPVOID                ace;
    ACE_HEADER            *aceHeader;
    ULONG                 i;

// XXX check return

    GetAclInformation (OldACL, (LPVOID) &aclSizeInfo, (DWORD) sizeof (aclSizeInfo), AclSizeInformation);

    //
    // Copy all of the ACEs to the new ACL
    //

    for (i = 0; i < aclSizeInfo.AceCount; i++)
    {
        //
        // Get the ACE and header info
        //

        if (!GetAce (OldACL, i, &ace))
            return HRESULT_FROM_WIN32 (GetLastError());

        aceHeader = (ACE_HEADER *) ace;

        //
        // Add the ACE to the new list
        //

        if (!AddAce (NewACL, ACL_REVISION, 0xffffffff, ace, aceHeader->AceSize))
		{
			ReportFailure("CopyACL", "AddAce", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (GetLastError());
		}
    }

    return ERROR_SUCCESS;
}

DWORD AddAccessDeniedACEToACL( PACL *Acl, DWORD PermissionMask, LPWSTR Principal )
{
    ACL_SIZE_INFORMATION  aclSizeInfo;
    int                   aclSize;
    DWORD                 returnValue;
    PSID                  principalSID;
    PACL                  oldACL, newACL;

    oldACL = *Acl;

    returnValue = GetPrincipalSID (Principal, &principalSID);
    if (returnValue != ERROR_SUCCESS)
        return returnValue;

// XXX check ret info

    GetAclInformation (oldACL, (LPVOID) &aclSizeInfo, (DWORD) sizeof (ACL_SIZE_INFORMATION), AclSizeInformation);

    aclSize = aclSizeInfo.AclBytesInUse +
              sizeof (ACL) + sizeof (ACCESS_DENIED_ACE) +
              GetLengthSid (principalSID) - sizeof (DWORD);

    newACL = (PACL) malloc (aclSize);

    if (!InitializeAcl (newACL, aclSize, ACL_REVISION2))
    {
        free (principalSID);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    if (!AddAccessDeniedAce (newACL, ACL_REVISION2, PermissionMask, principalSID))
    {
        free (principalSID);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    returnValue = CopyACL (oldACL, newACL);
    if (returnValue != ERROR_SUCCESS)
    {
        free (principalSID);
        return returnValue;
    }

    free (oldACL);
    *Acl = newACL;

    free (principalSID);
    return ERROR_SUCCESS;
}

DWORD AddAccessAllowedACEToACL( PACL *Acl, DWORD PermissionMask, LPWSTR Principal )
{
    ACL_SIZE_INFORMATION  aclSizeInfo;
    int                   aclSize;
    DWORD                 returnValue;
    PSID                  principalSID;
    PACL                  oldACL, newACL;

    oldACL = *Acl;

    returnValue = GetPrincipalSID (Principal, &principalSID);
    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("AddAccessAllowedACEToACL", "GetPrincipalSID", __FILE__, __LINE__);
        return returnValue;
	}


// XXX check ret info

    GetAclInformation (oldACL, (LPVOID) &aclSizeInfo, (DWORD) sizeof (ACL_SIZE_INFORMATION), AclSizeInformation);

    aclSize = aclSizeInfo.AclBytesInUse +
              sizeof (ACL) + sizeof (ACCESS_ALLOWED_ACE) +
              GetLengthSid (principalSID) - sizeof (DWORD);

    newACL = (PACL) malloc (aclSize);

    if (!InitializeAcl (newACL, aclSize, ACL_REVISION2))
    {
        free (principalSID);
		ReportFailure("AddAccessAllowedACEToACL", "InitializeAcl", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    returnValue = CopyACL (oldACL, newACL);
    if (returnValue != ERROR_SUCCESS)
    {
        free (principalSID);
		ReportFailure("AddAccessAllowedACEToACL", "CopyACL", __FILE__, __LINE__);
        return returnValue;
    }

    if (!AddAccessAllowedAce (newACL, ACL_REVISION2, PermissionMask, principalSID))
    {
        free (principalSID);
		ReportFailure("AddAccessAllowedACEToACL", "AddAccessAllowedAce", __FILE__, __LINE__);
        return HRESULT_FROM_WIN32 (GetLastError());
    }

    free (oldACL);
    *Acl = newACL;

    free (principalSID);
    return ERROR_SUCCESS;
}

DWORD RemovePrincipalFromACL( PACL Acl, LPWSTR Principal )
{
    ACL_SIZE_INFORMATION    aclSizeInfo;
    LONG                    i;
    LPVOID                  ace;
    ACCESS_ALLOWED_ACE      *accessAllowedAce;
    ACCESS_DENIED_ACE       *accessDeniedAce;
    SYSTEM_AUDIT_ACE        *systemAuditAce;
    PSID                    principalSID;
    DWORD                   returnValue;
    ACE_HEADER              *aceHeader;

    //
    // N.B. We go backwards because the numbering gets screwed up when
    // we delete ACEs; we would miss ACEs otherwise.
    //

    returnValue = GetPrincipalSID (Principal, &principalSID);
    if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("RemovePrincipalFromACL", "GetPrincipalSID", __FILE__, __LINE__);
		return returnValue;
	}


// XXX check ret info
    GetAclInformation (Acl, (LPVOID) &aclSizeInfo, (DWORD) sizeof (ACL_SIZE_INFORMATION), AclSizeInformation);

    for (i = aclSizeInfo.AceCount-1; i >= 0; i--)
    {
        if (!GetAce (Acl, i, &ace))
        {
            free (principalSID);
			ReportFailure("RemovePrincipalFromACL", "GetAce", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (GetLastError());
        }

        aceHeader = (ACE_HEADER *) ace;

        if (aceHeader->AceType == ACCESS_ALLOWED_ACE_TYPE)
        {
            accessAllowedAce = (ACCESS_ALLOWED_ACE *) ace;

            if (EqualSid (principalSID, (PSID) &accessAllowedAce->SidStart))
            {
                DeleteAce (Acl, i);
            }
        } else

        if (aceHeader->AceType == ACCESS_DENIED_ACE_TYPE)
        {
            accessDeniedAce = (ACCESS_DENIED_ACE *) ace;

            if (EqualSid (principalSID, (PSID) &accessDeniedAce->SidStart))
            {
                DeleteAce (Acl, i);
            }
        } else

        if (aceHeader->AceType == SYSTEM_AUDIT_ACE_TYPE)
        {
            systemAuditAce = (SYSTEM_AUDIT_ACE *) ace;

            if (EqualSid (principalSID, (PSID) &systemAuditAce->SidStart))
            {
                DeleteAce (Acl, i);
            }
        }
    }

    free (principalSID);
    return ERROR_SUCCESS;
}


BOOL
IsPrincipalInACL (
    PACL Acl,
    LPWSTR Principal
    )
{
    ACL_SIZE_INFORMATION    aclSizeInfo;
    LONG                    i;
    LPVOID                  ace;
    ACCESS_ALLOWED_ACE      *accessAllowedAce;
    ACCESS_DENIED_ACE       *accessDeniedAce;
    SYSTEM_AUDIT_ACE        *systemAuditAce;
    PSID                    principalSID;
    DWORD                   returnValue;
    ACE_HEADER              *aceHeader;

    returnValue = GetPrincipalSID (Principal, &principalSID);
    if (returnValue != ERROR_SUCCESS)
        return FALSE;

// XXX check ret info
    GetAclInformation (Acl, (LPVOID) &aclSizeInfo, (DWORD) sizeof (ACL_SIZE_INFORMATION), AclSizeInformation);

    for (i = aclSizeInfo.AceCount-1; i >= 0; i--)
    {
        if (!GetAce (Acl, i, &ace))
        {
            free (principalSID);
			ReportFailure("IsPrincipalInACL", "GetAce", __FILE__, __LINE__);
            return HRESULT_FROM_WIN32 (GetLastError());
        }

        aceHeader = (ACE_HEADER *) ace;

        if (aceHeader->AceType == ACCESS_ALLOWED_ACE_TYPE)
        {
            accessAllowedAce = (ACCESS_ALLOWED_ACE *) ace;

            if (EqualSid (principalSID, (PSID) &accessAllowedAce->SidStart))
            {
                free (principalSID);
                return TRUE;
            }
        } else

        if (aceHeader->AceType == ACCESS_DENIED_ACE_TYPE)
        {
            accessDeniedAce = (ACCESS_DENIED_ACE *) ace;

            if (EqualSid (principalSID, (PSID) &accessDeniedAce->SidStart))
            {
                free (principalSID);
                return TRUE;
            }
        } else

        if (aceHeader->AceType == SYSTEM_AUDIT_ACE_TYPE)
        {
            systemAuditAce = (SYSTEM_AUDIT_ACE *) ace;

            if (EqualSid (principalSID, (PSID) &systemAuditAce->SidStart))
            {
                free (principalSID);
                return TRUE;
            }
        }
    }

    free (principalSID);
    return FALSE;
}

